from lz4.frame import compress
from lz4.frame import decompress

from django.conf import settings

from django_valkey.compressors.base import BaseCompressor


class Lz4Compressor(BaseCompressor):
    """
    Lz4 compressor
    set with:
    CACHES = {
        "default": {
            # ...
            "OPTIONS": {
                "COMPRESSOR": "django_valkey.compressors.lz4.Lz4Compressor",
                }
            }
        }

    compression parameters:
    to set `compression_level` use `CACHE_COMPRESS_LEVEL` in your settings, defaults to 4.
    to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.
    to set `block_size` use `COMPRESS_LZ4_BLOCK_SIZE` in your settings, it defaults to 0.
    to set `content_checksum` use `COMPRESS_LZ4_CONTENT_CHECKSUM` in your settings, it defaults to 0.
    to set `blocked_linked` use `COMPRESS_LZ4_BLOCK_LINKED` in your settings, it defaults to True.
    to set `store_size` use `COMPRESS_LZ4_STORE_SIZE` in your settings, it defaults to True.
    """

    block_size = getattr(settings, "COMPRESS_LZ4_BLOCK_SIZE", 0)
    content_checksum = getattr(settings, "COMPRESS_LZ4_CONTENT_CHECKSUM", 0)
    block_linked = getattr(settings, "COMPRESS_LZ4_BLOCK_LINKED", True)
    store_size = getattr(settings, "COMPRESS_LZ4_STORE_SIZE", True)

    def _compress(self, value: bytes) -> bytes:
        return compress(
            value,
            compression_level=self.level or 0,
            block_size=self.block_size,
            content_checksum=self.content_checksum,
            block_linked=self.block_linked,
            store_size=self.store_size,
        )

    def _decompress(self, value: bytes) -> bytes:
        return decompress(value)



from typing import Any
from urllib.parse import urlparse, parse_qs

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from valkey.asyncio import Valkey as AValkey
from valkey.asyncio.connection import ConnectionPool, DefaultParser
from valkey.asyncio.sentinel import Sentinel
from valkey._parsers.url_parser import to_bool

from django_valkey.base_pool import BaseConnectionFactory


class AsyncConnectionFactory(BaseConnectionFactory[AValkey, ConnectionPool]):
    path_pool_cls = "valkey.asyncio.connection.ConnectionPool"
    path_base_cls = "valkey.asyncio.client.Valkey"

    async def disconnect(self, connection: type[AValkey]) -> None:
        await connection.connection_pool.disconnect()

    def get_parser_cls(self) -> type[DefaultParser] | type:
        cls = self.options.get("PARSER_CLS", None)
        if cls is None:
            return DefaultParser
        return import_string(cls)

    async def connect(self, url: str) -> AValkey | Any:
        params = self.make_connection_params(url)
        return await self.get_connection(params)

    async def get_connection(self, params: dict) -> AValkey | Any:
        pool = self.get_or_create_connection_pool(params)
        return await self.base_client_cls(
            connection_pool=pool, **self.base_client_cls_kwargs
        )


class AsyncSentinelConnectionFactory(AsyncConnectionFactory):
    def __init__(self, options: dict):
        try:
            self.sentinels = options["SENTINELS"]
        except KeyError:
            e = "SENTINELS must be provided as a list of (host, port)."
            raise ImproperlyConfigured(e)

        options.setdefault(
            "CONNECTION_POOL_CLASS", "valkey.asyncio.sentinel.SentinelConnectionPool"
        )
        super().__init__(options)
        connection_kwargs = self.make_connection_params(None)
        connection_kwargs.pop("url")
        connection_kwargs.update(self.pool_cls_kwargs)
        self._sentinel = Sentinel(
            self.sentinels,
            sentinel_kwargs=options.get("SENTINEL_KWARGS"),
            **connection_kwargs,
        )

    def get_connection_pool(self, params: dict) -> ConnectionPool | Any:
        url = urlparse(params["url"])
        cp_params = params
        cp_params.update(service_name=url.hostname, sentinel_manager=self._sentinel)
        pool = super().get_connection_pool(cp_params)

        is_master = parse_qs(url.query).get("is_master")
        if is_master:
            pool.is_master = to_bool(is_master[0])

        return pool



import contextlib
import functools
import logging
from asyncio import iscoroutinefunction
from typing import Any, TypeVar, Generic, Iterator, AsyncGenerator, Set, Callable

from django.conf import settings
from django.core.cache.backends.base import BaseCache
from django.utils.module_loading import import_string

from django_valkey.exceptions import ConnectionInterrupted

Client = TypeVar("Client")
Backend = TypeVar("Backend")


def omit_exception(method: Callable | None = None, return_value: Any | None = None):
    """
    Simple decorator that intercepts connection
    errors and ignores these if settings specify this.
    """

    if method is None:
        return functools.partial(omit_exception, return_value=return_value)

    @functools.wraps(method)
    def _decorator(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            if self._ignore_exceptions:
                if self._log_ignored_exceptions:
                    self.logger.exception("Exception ignored")

                return return_value
            raise e.__cause__  # noqa: B904

    @functools.wraps(method)
    async def _async_decorator(self, *args, **kwargs):
        try:
            return await method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            if self._ignore_exceptions:
                if self._log_ignored_exceptions:
                    self.logger.exception("Exception ignored")
                return return_value
            raise e.__cause__

    return _async_decorator if iscoroutinefunction(method) else _decorator


class BaseValkeyCache(BaseCache, Generic[Client, Backend]):
    def __init__(self, server: str, params: dict[str, Any]) -> None:
        super().__init__(params)
        self._server = server
        self._params = params
        self._default_scan_itersize = getattr(
            settings, "DJANGO_VALKEY_SCAN_ITERSIZE", 10
        )

        options: dict = params.get("OPTIONS", {})
        self._client_cls = self.get_client_class()
        self._client = None

        self._ignore_exceptions = options.get(
            "IGNORE_EXCEPTIONS",
            getattr(settings, "DJANGO_VALKEY_IGNORE_EXCEPTIONS", False),
        )
        self._log_ignored_exceptions = options.get(
            "LOG_IGNORE_EXCEPTIONS",
            getattr(settings, "DJANGO_VALKEY_LOG_IGNORED_EXCEPTIONS", False),
        )
        self.logger = (
            logging.getLogger(getattr(settings, "DJANGO_VALKEY_LOGGER", __name__))
            if self._log_ignored_exceptions
            else None
        )

    def get_client_class(self) -> type[Client] | type:
        options = self._params.get("OPTIONS", {})
        _client_cls = options.get("CLIENT_CLASS", self.DEFAULT_CLIENT_CLASS)
        return import_string(_client_cls)

    @property
    def client(self) -> Client:
        """
        Lazy client connection property.
        """
        if self._client is None:
            self._client = self._client_cls(self._server, self._params, self)
        return self._client

    @omit_exception
    def mset(self, *args, **kwargs):
        return self.client.mset(*args, **kwargs)

    @omit_exception
    async def amset(self, *args, **kwargs):
        return await self.client.amset(*args, **kwargs)

    @omit_exception
    def mget(self, *args, **kwargs):
        return self.client.mget(*args, **kwargs)

    @omit_exception
    async def amget(self, *args, **kwargs):
        return await self.client.amget(*args, **kwargs)

    @omit_exception
    def persist(self, *args, **kwargs) -> bool:
        return self.client.persist(*args, **kwargs)

    @omit_exception
    async def apersist(self, *args, **kwargs) -> bool:
        return await self.client.apersist(*args, **kwargs)

    @omit_exception
    def expire(self, *args, **kwargs) -> bool:
        return self.client.expire(*args, **kwargs)

    @omit_exception
    async def aexpire(self, *args, **kwargs) -> bool:
        return await self.client.aexpire(*args, **kwargs)

    @omit_exception
    def expire_at(self, *args, **kwargs) -> bool:
        return self.client.expire_at(*args, **kwargs)

    @omit_exception
    async def aexpire_at(self, *args, **kwargs) -> bool:
        return await self.client.aexpire_at(*args, **kwargs)

    @omit_exception
    def pexpire(self, *args, **kwargs) -> bool:
        return self.client.pexpire(*args, **kwargs)

    @omit_exception
    async def apexpire(self, *args, **kwargs) -> bool:
        return await self.client.apexpire(*args, **kwargs)

    @omit_exception
    def pexpire_at(self, *args, **kwargs) -> bool:
        return self.client.pexpire_at(*args, **kwargs)

    @omit_exception
    async def apexpire_at(self, *args, **kwargs) -> bool:
        return await self.client.apexpire_at(*args, **kwargs)

    @omit_exception
    def get_lock(self, *args, **kwargs):
        return self.client.get_lock(*args, **kwargs)

    lock = get_lock

    @omit_exception
    async def aget_lock(self, *args, **kwargs):
        return await self.client.aget_lock(*args, **kwargs)

    alock = aget_lock

    @omit_exception
    def delete_pattern(self, *args, **kwargs) -> int:
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return self.client.delete_pattern(*args, **kwargs)

    @omit_exception
    async def adelete_pattern(self, *args, **kwargs) -> int:
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return await self.client.adelete_pattern(*args, **kwargs)

    @omit_exception
    def ttl(self, *args, **kwargs) -> int | None:
        return self.client.ttl(*args, **kwargs)

    @omit_exception
    async def attl(self, *args, **kwargs) -> int:
        return await self.client.attl(*args, **kwargs)

    @omit_exception
    def pttl(self, *args, **kwargs) -> int:
        return self.client.pttl(*args, **kwargs)

    @omit_exception
    async def apttl(self, *args, **kwargs) -> int:
        return await self.client.apttl(*args, **kwargs)

    @omit_exception
    def iter_keys(self, *args, **kwargs) -> Iterator:
        return self.client.iter_keys(*args, **kwargs)

    @omit_exception
    async def aiter_keys(self, *args, **kwargs) -> AsyncGenerator:
        async with contextlib.aclosing(self.client.aiter_keys(*args, **kwargs)) as it:
            async for key in it:
                yield key

    @omit_exception
    def keys(self, *args, **kwargs) -> list[Any]:
        return self.client.keys(*args, **kwargs)

    @omit_exception
    async def akeys(self, *args, **kwargs) -> list[Any]:
        return await self.client.akeys(*args, **kwargs)

    @omit_exception
    def sadd(self, *args, **kwargs) -> int:
        return self.client.sadd(*args, **kwargs)

    @omit_exception
    async def asadd(self, *args, **kwargs) -> int:
        return await self.client.asadd(*args, **kwargs)

    @omit_exception
    def scard(self, *args, **kwargs) -> int:
        return self.client.scard(*args, **kwargs)

    @omit_exception
    async def ascard(self, *args, **kwargs) -> int:
        return await self.client.ascard(*args, **kwargs)

    @omit_exception
    def sdiff(self, *args, **kwargs) -> Set[Any]:
        return self.client.sdiff(*args, **kwargs)

    @omit_exception
    async def asdiff(self, *args, **kwargs) -> Set[Any]:
        return await self.client.asdiff(*args, **kwargs)

    @omit_exception
    def sdiffstore(self, *args, **kwargs) -> int:
        return self.client.sdiffstore(*args, **kwargs)

    @omit_exception
    async def asdiffstore(self, *args, **kwargs) -> int:
        return await self.client.asdiffstore(*args, **kwargs)

    @omit_exception
    def sinter(self, *args, **kwargs) -> Set[Any]:
        return self.client.sinter(*args, **kwargs)

    @omit_exception
    async def asinter(self, *args, **kwargs) -> Set[Any]:
        return await self.client.asinter(*args, **kwargs)

    @omit_exception
    def sinterstore(self, *args, **kwargs) -> int:
        return self.client.sinterstore(*args, **kwargs)

    @omit_exception
    async def asinterstore(self, *args, **kwargs) -> int:
        return await self.client.asinterstore(*args, **kwargs)

    @omit_exception
    def smismember(self, *args, **kwargs) -> list[bool]:
        return self.client.smismember(*args, **kwargs)

    @omit_exception
    async def asmismember(self, *args, **kwargs) -> list[bool]:
        return await self.client.asmismember(*args, **kwargs)

    @omit_exception
    def sismember(self, *args, **kwargs) -> bool:
        return self.client.sismember(*args, **kwargs)

    @omit_exception
    async def asismember(self, *args, **kwargs) -> bool:
        return await self.client.asismember(*args, **kwargs)

    @omit_exception
    def smembers(self, *args, **kwargs) -> Set[Any]:
        return self.client.smembers(*args, **kwargs)

    @omit_exception
    async def asmembers(self, *args, **kwargs) -> Set[Any]:
        return await self.client.asmembers(*args, **kwargs)

    @omit_exception
    def smove(self, *args, **kwargs) -> bool:
        return self.client.smove(*args, **kwargs)

    @omit_exception
    async def asmove(self, *args, **kwargs) -> bool:
        return await self.client.asmove(*args, **kwargs)

    @omit_exception
    def spop(self, *args, **kwargs) -> Set | Any:
        return self.client.spop(*args, **kwargs)

    @omit_exception
    async def aspop(self, *args, **kwargs) -> Set | Any:
        return await self.client.aspop(*args, **kwargs)

    @omit_exception
    def srandmember(self, *args, **kwargs) -> list | Any:
        return self.client.srandmember(*args, **kwargs)

    @omit_exception
    async def asrandmember(self, *args, **kwargs) -> list | Any:
        return await self.client.asrandmember(*args, **kwargs)

    @omit_exception
    def srem(self, *args, **kwargs) -> int:
        return self.client.srem(*args, **kwargs)

    @omit_exception
    async def asrem(self, *args, **kwargs) -> int:
        return await self.client.asrem(*args, **kwargs)

    @omit_exception
    def sscan(self, *args, **kwargs) -> Set[Any]:
        return self.client.sscan(*args, **kwargs)

    @omit_exception
    async def asscan(self, *args, **kwargs) -> Set[Any]:
        return await self.client.asscan(*args, **kwargs)

    @omit_exception
    def sscan_iter(self, *args, **kwargs) -> Iterator:
        return self.client.sscan_iter(*args, **kwargs)

    @omit_exception
    async def asscan_iter(self, *args, **kwargs) -> AsyncGenerator:
        async with contextlib.aclosing(self.client.asscan_iter(*args, **kwargs)) as it:
            async for key in it:
                yield key

    @omit_exception
    def sunion(self, *args, **kwargs) -> Set[Any]:
        return self.client.sunion(*args, **kwargs)

    @omit_exception
    async def asunion(self, *args, **kwargs) -> Set[Any]:
        return await self.client.asunion(*args, **kwargs)

    @omit_exception
    def sunionstore(self, *args, **kwargs) -> int:
        return self.client.sunionstore(*args, **kwargs)

    @omit_exception
    async def asunionstore(self, *args, **kwargs) -> int:
        return await self.client.asunionstore(*args, **kwargs)

    @omit_exception
    def hset(self, *args, **kwargs) -> int:
        return self.client.hset(*args, **kwargs)

    @omit_exception
    async def ahset(self, *args, **kwargs) -> int:
        return await self.client.hset(*args, **kwargs)

    @omit_exception
    def hdel(self, *args, **kwargs) -> int:
        return self.client.hdel(*args, **kwargs)

    @omit_exception
    async def ahdel(self, *args, **kwargs) -> int:
        return await self.client.hdel(*args, **kwargs)

    @omit_exception
    def hlen(self, *args, **kwargs) -> int:
        return self.client.hlen(*args, **kwargs)

    @omit_exception
    async def ahlen(self, *args, **kwargs) -> int:
        return await self.client.hlen(*args, **kwargs)

    @omit_exception
    def hkeys(self, *args, **kwargs) -> list[Any]:
        return self.client.hkeys(*args, **kwargs)

    @omit_exception
    async def ahkeys(self, *args, **kwargs) -> list[Any]:
        return await self.client.ahkeys(*args, **kwargs)

    @omit_exception
    def hexists(self, *args, **kwargs) -> bool:
        return self.client.hexists(*args, **kwargs)

    @omit_exception
    async def ahexists(self, *args, **kwargs) -> bool:
        return await self.client.ahexists(*args, **kwargs)



import lzma

from django.conf import settings

from django_valkey.compressors.base import BaseCompressor


class LzmaCompressor(BaseCompressor):
    """
    LzmaCompressor
    set with:
    CACHES = {
        "default": {
            # ...
            "OPTIONS": {
                "COMPRESSOR": "django_valkey.compressors.lzma.LzmaCompressor",
                }
            }
        }

    compression parameters:
    to set `preset` use `CACHE_COMPRESS_LEVEL` in your settings, defaults to 4.
    to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.
    to set `format` use `COMPRESS_LZMA_FORMAT` in your settings, it defaults to 1.
    to set `check` use `COMPRESS_LZMA_CHECK` in your settings, it defaults to -1.
    to set `filters` use `COMPRESS_LZMA_FILTERS` in your settings, it defaults to None.

    decompression parameters:
    to set `memlimit` use `DECOMPRESS_LZMA_MEMLIMIT` in your settings, it defaults to None.
    to set `format` use `DECOMPRESS_LZMA_FORMAT` in your settings, it defaults to 0.
    to set `filters` use `DECOMPRESS_LZMA_FILTERS` in your settings, it defaults to None.
    """

    format = getattr(settings, "COMPRESS_LZMA_FORMAT", 1)
    check = getattr(settings, "COMPRESS_LZMA_CHECK", -1)
    filters: dict | None = getattr(settings, "COMPRESS_LZMA_FILTERS", None)

    memlimit: int = getattr(settings, "DECOMPRESS_LZMA_MEMLIMIT", None)
    decomp_format = getattr(settings, "DECOMPRESS_LZMA_FORMAT", 0)
    decomp_filters = getattr(settings, "DECOMPRESS_LZMA_FILTERS", None)

    def _compress(self, value: bytes) -> bytes:
        return lzma.compress(
            value,
            preset=self.level or 4,
            format=self.format or 4,
            check=self.check,
            filters=self.filters,
        )

    def _decompress(self, value: bytes) -> bytes:
        return lzma.decompress(
            value,
            memlimit=self.memlimit,
            format=self.decomp_format,
            filters=self.decomp_filters,
        )



from typing import TypeVar, Generic, Any

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

Pool = TypeVar("Pool")
Base = TypeVar("Base")


class BaseConnectionFactory(Generic[Base, Pool]):
    # Store connection pool by cache backend options.
    #
    # _pools is a process-global, as otherwise _pools is cleared every time
    # ConnectionFactory is instantiated, as Django creates new cache client
    # (DefaultClient) instance for every request.

    _pools: dict[str, Pool | Any] = {}

    def __init__(self, options: dict):
        pool_cls_path = options.get("CONNECTION_POOL_CLASS", self.path_pool_cls)
        self.pool_cls: type[Pool] | type = import_string(pool_cls_path)
        self.pool_cls_kwargs = options.get("CONNECTION_POOL_KWARGS", {})

        base_client_cls_path = options.get("BASE_CLIENT_CLASS", self.path_base_cls)
        self.base_client_cls: type[Base] | type = import_string(base_client_cls_path)
        self.base_client_cls_kwargs = options.get("BASE_CLIENT_KWARGS", {})

        self.options = options

    def make_connection_params(self, url: str | None) -> dict:
        """
        Given a main connection parameters, build a complete
        dict of connection parameters.
        """

        kwargs = {
            "url": url,
            "parser_class": self.get_parser_cls(),
        }

        socket_timeout = self.options.get("SOCKET_TIMEOUT", None)
        # TODO: do we need to check for existence?
        if socket_timeout:
            if not isinstance(socket_timeout, (int, float)):
                error_message = "Socket timeout should be float or integer"
                raise ImproperlyConfigured(error_message)
            kwargs["socket_timeout"] = socket_timeout

        socket_connect_timeout = self.options.get("SOCKET_CONNECT_TIMEOUT", None)
        if socket_connect_timeout:
            if not isinstance(socket_connect_timeout, (int, float)):
                error_message = "Socket connect timeout should be float or integer"
                raise ImproperlyConfigured(error_message)
            kwargs["socket_connect_timeout"] = socket_connect_timeout

        password = self.options.get("PASSWORD", None)
        if password:
            kwargs["password"] = password

        return kwargs

    def get_connection_pool(self, params: dict) -> Pool | Any:
        """
        Given a connection parameters, return a new
        connection pool for them.

        Overwrite this method if you want a custom
        behavior on creating connection pool.
        """
        cp_params = params
        cp_params.update(self.pool_cls_kwargs)
        pool = self.pool_cls.from_url(**cp_params)

        if pool.connection_kwargs.get("password", None) is None:
            pool.connection_kwargs["password"] = params.get("password", None)
            pool.reset()

        return pool

    def get_or_create_connection_pool(self, params: dict) -> Pool | Any:
        """
        Given a connection parameters and return a new
        or cached connection pool for them.

        Reimplement this method if you want distinct
        connection pool instance caching behavior.
        """
        key: str = params["url"]
        if key not in self._pools:
            self._pools[key] = self.get_connection_pool(params)
        return self._pools[key]

    def get_connection(self, params: dict) -> Base | Any:
        """
        Given a now preformatted params, return a
        new connection.

        The default implementation uses a cached pools
        for create new connection.
        """
        raise NotImplementedError

    def connect(self, url: str) -> Base | Any:
        """
        Given a basic connection parameters,
        return a new connection.
        """
        raise NotImplementedError

    def disconnect(self, connection: type[Base]):
        raise NotImplementedError

    def get_parser_cls(self):
        raise NotImplementedError



from django.conf import settings

from django_valkey.exceptions import CompressorError


class BaseCompressor:
    min_length = getattr(settings, "CACHE_COMPRESS_MIN_LENGTH", 15)
    level: int | None = getattr(settings, "CACHE_COMPRESS_LEVEL", None)

    def __init__(self, options):
        self._options: dict = options

    def compress(self, value):
        if len(value) > self.min_length:
            return self._compress(value)
        return value

    def _compress(self, value: bytes) -> bytes:
        raise NotImplementedError

    def decompress(self, value: bytes) -> bytes:
        try:
            return self._decompress(value)
        except Exception as e:
            raise CompressorError from e

    def _decompress(self, value: bytes) -> bytes:
        raise NotImplementedError



import bz2

from django_valkey.compressors.base import BaseCompressor


class Bz2Compressor(BaseCompressor):
    """
    Bz2Compressor
    set with:
    CACHES = {
        "default": {
            # ...
            "OPTIONS": {
                "COMPRESSOR": "django_valkey.compressors.bz2.Bz2Compressor",
                }
            }
        }

    compression parameters:
    to set `compresslevel` use `CACHE_COMPRESS_LEVEL` in your settings, defaults to 4.
    to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.
    """

    def _compress(self, value: bytes) -> bytes:
        return bz2.compress(value, compresslevel=self.level or 9)

    def _decompress(self, value: bytes) -> bytes:
        return bz2.decompress(value)



from django_valkey.async_cache.client.default import AsyncDefaultClient
from django_valkey.async_cache.client.herd import AsyncHerdClient
from django_valkey.async_cache.client.sentinel import AsyncSentinelClient

__all__ = ["AsyncDefaultClient", "AsyncHerdClient", "AsyncSentinelClient"]



from typing import Any

from django.utils.module_loading import import_string
from valkey.cluster import ValkeyCluster
from valkey.connection import ConnectionPool, DefaultParser

from django_valkey.base_pool import BaseConnectionFactory


class ClusterConnectionFactory(BaseConnectionFactory[ValkeyCluster, ConnectionPool]):
    path_pool_cls = "valkey.connection.ConnectionPool"
    path_base_cls = "valkey.cluster.ValkeyCluster"

    def disconnect(self, connection: ValkeyCluster) -> None:
        connection.disconnect_connection_pools()

    def get_parser_cls(self):
        cls = self.options.get("PARSER_CLS", None)
        if cls is None:
            return DefaultParser
        return import_string(cls)

    def connect(self, url: str) -> ValkeyCluster:
        params = self.make_connection_params(url)
        return self.get_connection(params)

    def get_connection(self, params: dict) -> ValkeyCluster | Any:
        return self.base_client_cls(
            url=params["url"],
            parser_class=params["parser_class"],
            **self.base_client_cls_kwargs,
        )



from valkey import Valkey

from django_valkey.base import BaseValkeyCache, omit_exception
from django_valkey.client import DefaultClient

CONNECTION_INTERRUPTED = object()


class ValkeyCache(BaseValkeyCache[DefaultClient, Valkey]):
    DEFAULT_CLIENT_CLASS = "django_valkey.client.DefaultClient"

    @omit_exception
    def set(self, *args, **kwargs):
        return self.client.set(*args, **kwargs)

    @omit_exception
    def incr_version(self, *args, **kwargs):
        return self.client.incr_version(*args, **kwargs)

    @omit_exception
    def add(self, *args, **kwargs):
        return self.client.add(*args, **kwargs)

    def get(self, key, default=None, version=None, client=None):
        value = self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default
        return value

    @omit_exception(return_value=CONNECTION_INTERRUPTED)
    def _get(self, key, default, version, client):
        return self.client.get(key, default=default, version=version, client=client)

    @omit_exception
    def delete(self, *args, **kwargs):
        result = self.client.delete(*args, **kwargs)
        return bool(result)

    @omit_exception
    def delete_many(self, *args, **kwargs):
        return self.client.delete_many(*args, **kwargs)

    @omit_exception
    def clear(self):
        return self.client.clear()

    @omit_exception(return_value={})
    def get_many(self, *args, **kwargs):
        return self.client.get_many(*args, **kwargs)

    @omit_exception
    def set_many(self, *args, **kwargs):
        return self.client.set_many(*args, **kwargs)

    @omit_exception
    def incr(self, *args, **kwargs):
        return self.client.incr(*args, **kwargs)

    @omit_exception
    def decr(self, *args, **kwargs):
        return self.client.decr(*args, **kwargs)

    @omit_exception
    def has_key(self, *args, **kwargs):
        return self.client.has_key(*args, **kwargs)

    @omit_exception
    def keys(self, *args, **kwargs):
        return self.client.keys(*args, **kwargs)

    @omit_exception
    def iter_keys(self, *args, **kwargs):
        return self.client.iter_keys(*args, **kwargs)

    @omit_exception
    def close(self):
        self.client.close()

    @omit_exception
    def touch(self, *args, **kwargs):
        return self.client.touch(*args, **kwargs)



from valkey.asyncio.client import Valkey as AValkey

from django_valkey.base import BaseValkeyCache
from django_valkey.cache import omit_exception, CONNECTION_INTERRUPTED
from django_valkey.async_cache.client.default import AsyncDefaultClient


class AsyncValkeyCache(BaseValkeyCache[AsyncDefaultClient, AValkey]):
    DEFAULT_CLIENT_CLASS = "django_valkey.async_cache.client.default.AsyncDefaultClient"

    mset = BaseValkeyCache.amset

    keys = BaseValkeyCache.akeys

    iter_keys = BaseValkeyCache.aiter_keys

    ttl = BaseValkeyCache.attl

    pttl = BaseValkeyCache.apttl

    persist = BaseValkeyCache.apersist

    expire = BaseValkeyCache.aexpire

    expire_at = BaseValkeyCache.aexpire_at

    pexpire = BaseValkeyCache.apexpire

    pexpire_at = BaseValkeyCache.apexpire_at

    lock = alock = get_lock = BaseValkeyCache.aget_lock

    sadd = BaseValkeyCache.asadd

    scard = BaseValkeyCache.ascard

    sdiff = BaseValkeyCache.asdiff

    sdiffstore = BaseValkeyCache.asdiffstore

    sinter = BaseValkeyCache.asinter

    sinterstore = BaseValkeyCache.asinterstore

    sismember = BaseValkeyCache.asismember

    smembers = BaseValkeyCache.asmembers

    smove = BaseValkeyCache.asmove

    spop = BaseValkeyCache.aspop

    srandmember = BaseValkeyCache.asrandmember

    srem = BaseValkeyCache.asrem

    sscan = BaseValkeyCache.asscan

    sscan_iter = BaseValkeyCache.asscan_iter

    smismember = BaseValkeyCache.asmismember

    sunion = BaseValkeyCache.asunion

    sunionstore = BaseValkeyCache.asunionstore

    hset = BaseValkeyCache.ahset

    hdel = BaseValkeyCache.ahdel

    hlen = BaseValkeyCache.ahlen

    hkeys = BaseValkeyCache.ahkeys

    hexists = BaseValkeyCache.ahexists

    @omit_exception
    async def set(self, *args, **kwargs):
        return await self.client.aset(*args, **kwargs)

    aset = set

    @omit_exception
    async def incr_version(self, *args, **kwargs):
        return await self.client.aincr_version(*args, **kwargs)

    aincr_version = incr_version

    @omit_exception
    async def add(self, *args, **kwargs):
        return await self.client.aadd(*args, **kwargs)

    aadd = add

    async def get(self, key, default=None, version=None, client=None):
        value = await self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default
        return value

    aget = get

    @omit_exception(return_value=CONNECTION_INTERRUPTED)
    async def _get(self, key, default=None, version=None, client=None):
        return await self.client.aget(key, default, version, client)

    async def delete(self, *args, **kwargs):
        result = await self.client.adelete(*args, **kwargs)
        return bool(result)

    adelete = delete

    @omit_exception
    async def delete_many(self, *args, **kwargs):
        return await self.client.adelete_many(*args, **kwargs)

    adelete_many = delete_many

    @omit_exception
    async def clear(self):
        return await self.client.aclear()

    aclear = clear

    @omit_exception(return_value={})
    async def get_many(self, *args, **kwargs):
        return await self.client.aget_many(*args, **kwargs)

    aget_many = get_many

    @omit_exception
    async def set_many(self, *args, **kwargs):
        return await self.client.aset_many(*args, **kwargs)

    aset_many = set_many

    @omit_exception
    async def incr(self, *args, **kwargs):
        return await self.client.aincr(*args, **kwargs)

    aincr = incr

    @omit_exception
    async def decr(self, *args, **kwargs):
        return await self.client.adecr(*args, **kwargs)

    adecr = decr

    @omit_exception
    async def has_key(self, *args, **kwargs):
        return await self.client.ahas_key(*args, **kwargs)

    ahas_key = has_key

    @omit_exception
    async def aclose(self, *args, **kwargs):
        return await self.client.aclose()

    @omit_exception
    async def touch(self, *args, **kwargs):
        return await self.client.touch(*args, **kwargs)

    atouch = touch



from valkey.cluster import ValkeyCluster

from django_valkey.base import BaseValkeyCache
from django_valkey.cache import CONNECTION_INTERRUPTED
from django_valkey.cluster_cache.client import DefaultClusterClient


class ClusterValkeyCache(BaseValkeyCache[DefaultClusterClient, ValkeyCluster]):
    DEFAULT_CLIENT_CLASS = "django_valkey.cluster_cache.client.DefaultClusterClient"

    def set(self, *args, **kwargs):
        return self.client.set(*args, **kwargs)

    def incr(self, *args, **kwargs):
        return self.client.incr(*args, **kwargs)

    def decr(self, *args, **kwargs):
        return self.client.decr(*args, **kwargs)

    def incr_version(self, *args, **kwargs):
        return self.client.incr_version(*args, **kwargs)

    def add(self, *args, **kwargs):
        return self.client.add(*args, **kwargs)

    def get(self, key, default=None, version=None, client=None):
        value = self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default

        return value

    def _get(self, key, default=None, version=None, client=None):
        return self.client.get(key, default, version, client)

    def delete(self, *args, **kwargs):
        result = self.client.delete(*args, **kwargs)
        return bool(result)

    def delete_pattern(self, *args, **kwargs):
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return self.client.delete_pattern(*args, **kwargs)

    def delete_many(self, *args, **kwargs):
        return self.client.delete_many(*args, **kwargs)

    def clear(self):
        return self.client.clear()

    def get_many(self, *args, **kwargs):
        return self.client.get_many(*args, **kwargs)

    def mget(self, *args, **kwargs):
        return self.client.mget(*args, **kwargs)

    def mget_nonatomic(self, *args, **kwargs):
        return self.client.mget_nonatomic(*args, **kwargs)

    def set_many(self, *args, **kwargs):
        return self.client.set_many(*args, **kwargs)

    def mset(self, *args, **kwargs):
        return self.client.mset(*args, **kwargs)

    def mset_nonatomic(self, *args, **kwargs):
        return self.client.mset_nonatomic(*args, **kwargs)

    def close(self):
        self.client.close()

    def has_key(self, *args, **kwargs):
        return self.client.has_key(*args, **kwargs)

    def touch(self, *args, **kwargs):
        return self.client.touch(*args, **kwargs)



VERSION = (0, 2, 1)
__version__ = ".".join(map(str, VERSION))


def get_valkey_connection(alias="default", write=True):
    """
    Helper used for obtaining a raw valkey client.
    """

    from django.core.cache import caches

    cache = caches[alias]

    error_message = "This backend does not support this feature"
    if not hasattr(cache, "client"):
        raise NotImplementedError(error_message)

    if not hasattr(cache.client, "get_client"):
        raise NotImplementedError(error_message)

    return cache.client.get_client(write)



import pickle
from typing import Any

from django.core.exceptions import ImproperlyConfigured

from django_valkey.serializers.base import BaseSerializer


class PickleSerializer(BaseSerializer):
    def __init__(self, options) -> None:
        self._pickle_version = pickle.DEFAULT_PROTOCOL
        self.setup_pickle_version(options)

        super().__init__(options=options)

    def setup_pickle_version(self, options) -> None:
        if "PICKLE_VERSION" in options:
            try:
                self._pickle_version = int(options["PICKLE_VERSION"])
                if self._pickle_version > pickle.HIGHEST_PROTOCOL:
                    error_message = (
                        f"PICKLE_VERSION can't be higher than pickle.HIGHEST_PROTOCOL:"
                        f" {pickle.HIGHEST_PROTOCOL}"
                    )
                    raise ImproperlyConfigured(error_message)
            except (ValueError, TypeError) as e:
                error_message = "PICKLE_VERSION value must be an integer"
                raise ImproperlyConfigured(error_message) from e

    def dumps(self, value: Any) -> bytes:
        return pickle.dumps(value, self._pickle_version)

    def loads(self, value: bytes) -> Any:
        return pickle.loads(value)



from typing import Dict

from valkey.cluster import ValkeyCluster
from valkey.typing import KeyT, EncodableT

from django_valkey.base_client import BaseClient, _main_exceptions
from django_valkey.exceptions import ConnectionInterrupted


class DefaultClusterClient(BaseClient[ValkeyCluster]):
    CONNECTION_FACTORY_PATH = (
        "django_valkey.cluster_cache.pool.ClusterConnectionFactory"
    )

    def readonly(self, target_nodes=None, client=None):
        client = self._get_client(write=True, client=client)
        return client.readonly(target_nodes)

    def readwrite(self, target_nodes=None, client=None):
        client = self._get_client(write=True, client=client)
        return client.readwrite(target_nodes)

    def keys(
        self,
        pattern="*",
        target_nodes=ValkeyCluster.DEFAULT_NODE,
        version=None,
        client=None,
        **kwargs,
    ):
        client = self._get_client(client=client)
        pattern = self.make_pattern(pattern, version=version)

        try:
            keys = client.keys(pattern=pattern, target_nodes=target_nodes, **kwargs)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return {self.reverse_key(key.decode()) for key in keys}

    def mset(
        self,
        data: Dict[KeyT, EncodableT],
        version=None,
        client=None,
        nx=False,
        atomic=True,
    ) -> None:
        """
        Access valkey's mset method.
        it is important to take care of cluster limitations mentioned here: https://valkey-py.readthedocs.io/en/latest/clustering.html#multi-key-commands
        """
        data = {
            self.make_key(k, version=version): self.encode(v) for k, v in data.items()
        }
        client = self._get_client(write=True, client=client)
        if not atomic:
            return client.mset_nonatomic(data)
        if nx:
            return client.msetnx(data)
        try:
            return client.mset(data)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    set_many = mset

    def msetnx(self, data: Dict[KeyT, EncodableT], version=None, client=None):
        try:
            return self.mset(data, version=version, client=client, nx=True)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def mset_nonatomic(self, data: Dict[KeyT, EncodableT], version=None, client=None):
        try:
            return self.mset(data, version=version, client=client, atomic=False)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    set_many = mset_nonatomic

    def mget_nonatomic(self, keys, version=None, client=None):
        client = self._get_client(client=client)
        map_keys = {self.make_key(k, version=version): k for k in keys}
        try:
            values = client.mget_nonatomic(map_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        recovered_data = {}
        for key, value in zip(keys, values):
            if value is None:
                continue
            recovered_data[map_keys[key]] = self.decode(value)
        return recovered_data

    def keyslot(self, key, version=None, client=None):
        client = self._get_client(client=client)
        key = self.make_key(key, version=version)
        return client.keyslot(key)

    def flush_cache(self, client=None):
        client = self._get_client(client=client)
        return client.flush_cache()

    def invalidate_key_from_cache(self, client=None):
        client = self._get_client(client=client)
        return client.invalidate_key_from_cache()



from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from django.core.exceptions import ImproperlyConfigured
from valkey.sentinel import SentinelConnectionPool

from django_valkey.client.default import DefaultClient


def replace_query(url, query):
    return urlunparse((*url[:4], urlencode(query, doseq=True), url[5]))


class SentinelClient(DefaultClient):
    CONNECTION_FACTORY_PATH = "django_valkey.pool.SentinelConnectionFactory"

    """
    Sentinel client which uses the single valkey URL specified by the CACHE's
    LOCATION to create a LOCATION configuration for two connection pools; One
    pool for the primaries and another pool for the replicas, and upon
    connecting ensures the connection pool factory is configured correctly.
    """

    def __init__(self, server, params, backend):
        if isinstance(server, str):
            url = urlparse(server)
            primary_query = parse_qs(url.query, keep_blank_values=True)
            replica_query = primary_query
            primary_query["is_master"] = [1]  # type: ignore
            replica_query["is_master"] = [0]  # type: ignore

            server = [replace_query(url, i) for i in (primary_query, replica_query)]

        super().__init__(server, params, backend)

    def connect(self, *args, **kwargs):
        connection = super().connect(*args, **kwargs)
        if not isinstance(connection.connection_pool, SentinelConnectionPool):
            error_message = (
                "Settings DJANGO_VALKEY_CONNECTION_FACTORY or "
                "CACHE[].OPTIONS.CONNECTION_POOL_CLASS is not configured correctly."
            )
            raise ImproperlyConfigured(error_message)

        return connection



class CacheKey(str):
    """
    A stub string class that we can use to check if a key was created already.
    """

    def original_key(self) -> str:
        return self.rsplit(":", 1)[1]


def default_reverse_key(key: str) -> str:
    return key.split(":", 2)[2]



import socket
import time
from typing import Tuple, Any

from django.conf import settings
from valkey import Valkey
from valkey.exceptions import ConnectionError, ResponseError, TimeoutError
from valkey.typing import KeyT, EncodableT

from django_valkey.async_cache.client import AsyncDefaultClient
from django_valkey.base_client import DEFAULT_TIMEOUT
from django_valkey.client.herd import Marker, _is_expired
from django_valkey.exceptions import ConnectionInterrupted

_main_exceptions = (ConnectionError, ResponseError, TimeoutError, socket.timeout)


class AsyncHerdClient(AsyncDefaultClient):
    def __init__(self, *args, **kwargs):
        self._marker = Marker()
        self._herd_timeout: int = getattr(settings, "CACHE_HERD_TIMEOUT", 60)
        super().__init__(*args, **kwargs)

    async def _pack(self, value: Any, timeout) -> Tuple[Marker, Any, int]:
        herd_timeout = (timeout or self._backend.default_timeout) + int(time.time())
        return self._marker, value, herd_timeout

    async def _unpack(self, value: Tuple[Marker, Any, int]) -> Tuple[Any, bool]:
        try:
            marker, unpacked, herd_timeout = value
        except (ValueError, TypeError):
            return value, False

        if not isinstance(marker, Marker):
            return value, False

        now = time.time()
        if herd_timeout < now:
            x = now - herd_timeout
            return unpacked, _is_expired(x, self._herd_timeout)

        return unpacked, False

    async def set(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | None = None,
        nx: bool = False,
        xx: bool = False,
    ):
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        if timeout is None or timeout <= 0:
            return await super().aset(
                key,
                value,
                timeout=timeout,
                version=version,
                client=client,
                nx=nx,
                xx=xx,
            )

        packed = await self._pack(value, timeout)
        real_timeout = timeout + self._herd_timeout

        return await super().aset(
            key,
            packed,
            timeout=real_timeout,
            version=version,
            client=client,
            nx=nx,
            xx=xx,
        )

    aset = set

    async def get(self, key, default=None, version=None, client=None):
        packed = await super().aget(
            key, default=default, version=version, client=client
        )
        val, refresh = await self._unpack(packed)

        if refresh:
            return default

        return val

    aget = get

    async def get_many(self, keys, version=None, client=None):
        client = await self._get_client(write=False, client=client)

        if not keys:
            return {}

        recovered_data = {}

        new_keys = [await self.make_key(key, version=version) for key in keys]
        map_keys = dict(zip(new_keys, keys))

        try:
            pipeline = await client.pipeline()
            for key in new_keys:
                await pipeline.get(key)
            results = await pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(new_keys, results):
            if value is None:
                continue

            val, refresh = await self._unpack(await self.decode(value))
            recovered_data[map_keys[key]] = None if refresh else val

        return recovered_data

    aget_many = get_many

    async def mget(self, keys, version=None, client=None):
        if not keys:
            return {}

        client = await self._get_client(write=False, client=client)

        recovered_data = {}

        new_keys = [await self.make_key(key, version=version) for key in keys]
        map_keys = dict(zip(new_keys, keys))

        try:
            results = await client.mget(new_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(new_keys, results):
            if value is None:
                continue

            val, refresh = await self._unpack(await self.decode(value))
            recovered_data[map_keys[key]] = None if refresh else val

        return recovered_data

    amget = mget

    async def set_many(
        self, data, timeout=DEFAULT_TIMEOUT, version=None, client=None, herd=True
    ):
        """
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        client = await self._get_client(write=True, client=client)

        set_function = self.aset if herd else super().aset

        try:
            pipeline = await client.pipeline()
            for key, value in data.items():
                await set_function(
                    key, value, timeout, version=version, client=pipeline
                )
            await pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    ast_many = set_many

    def incr(self, *args, **kwargs):
        raise NotImplementedError

    aincr = incr

    def decr(self, *args, **kwargs):
        raise NotImplementedError

    adecr = decr

    async def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None, client=None):
        client = await self._get_client(write=True, client=client)

        value = await self.aget(key, version=version, client=client)
        if value is None:
            return False

        await self.aset(key, value, timeout=timeout, version=version, client=client)
        return True

    atouch = touch
